/**
 * var iterate = require('lib/IterateOverPullFilesAndDoSomething');
 * iterate('absolute path to file exporting function');
 *
 * The function should do all work and return either a promise or a value.
 * Any asynchronous work should return a promise that is resolved when that work
 * is done.
 * Synchronous work can return anything not thenable.
 *
 * Optionally include the last known pull Id to be completed. If this field
 * isn't present, start from the beginning.
 *
 * Environment variables:
 * PULLS_PATH : path to the directory of pulls, default '/mnt/flashdrive/pulls/'
 * WORKERS : number of worker processes to use, default 4
 * PULLS_ACTIVE_PER_WORKER : number of pulls to actively be working on per
 *   worker, default 200
 * PULLS_COMPLETE_PER_WORKER : number of pulls to complete in each worker before
 *   cycling out the worker process for a new one. default 1000
 * MAX_PULL_ID : ...
 *
 * From wherever you run the file, you can `watch -n 1 cat status.txt`
 * to see the status of the iteration.
 */

module.exports = function (pathOfModuleExportingFunction, lastPullId) {
  var main = new Main(processes, workerPath, pathOfModuleExportingFunction,
      lastPullId || -1);
  main.run();
};

var Parent = require('lib/Parent'),
    Promise = require('promise'),
    _ = require('lodash'),
    fs = require('fs'),
    processes = process.env.WORKERS || 3,
    workerPath = __dirname + '/IterateOverPullFilesWorker.js',
    troublesomePullIdsLoc = '../troublesomePullIds.json',
    troublesomePullIds = require(troublesomePullIdsLoc),
    debug = require('debug')('ParseForComments'),
    promiseWriteFile = Promise.denodeify(fs.writeFile.bind(fs));

// select min/max(id) from pull_requests;
// var minPullId = 6350
var maxPullId = process.env.MAX_PULL_ID || 6909038;


function Main() {
  Parent.apply(this, arguments);
  this.pastCompletions = [];
  this.completions = [];
  this.completeUpTos = [];
  this.startTime = Date.now();
  this.intervals = [];
  this.intervals.push(setInterval(this.updateDisplay.bind(this), 1000));
  this.callbackPath = arguments[2];
  this.lastPullId = arguments[3];
}

Main.prototype = Object.create(Parent.prototype);

Main.prototype.onStop = function onStop() {
  _.each(this.intervals, clearInterval);
  this.updateDisplay(true);
  Parent.prototype.onStop.apply(this, arguments);
};

Main.prototype.calculateTotalCompletions =
    function calculateTotalCompletions() {
  var completions = _.reduce(this.pastCompletions, sum, 0);
  completions += _.reduce(this.completions, sum, 0);

  return completions;

  function sum(total, num) {
    return total + (num % 1 !== 0 ? 0 : num);
  };
}

Main.prototype.calculateAvgRate = function calculateAverageRate() {
  return this.calculateTotalCompletions() /
      (Date.now() - this.startTime) * 1000 * 60;
};

Main.prototype.calculateComleteUpto = function completeUpTo() {
  // if we're finished and all of the childen are stopped,
  // use the largest completeUpTo value
  if (this.lastPullId > maxPullId &&
      _.every(this.children, function(child) {
        return !child.running;
      })) {
    // it went over in the loop
    return this.lastPullId - 1;
  } else {
    return _.reduce(this.completeUpTos, function(min, num) {
      return num ? (num < min ? num : min) : 0;
    }, maxPullId);
  }
};

Main.prototype.onChildMessage = function onChildMessage(child, message) {
  switch (message.code) {
  case 'troublesomePullId':
    debug('%s troublesomePullId: %s', child.index, message.id);

    troublesomePullIds.push(message.id);
    promiseWriteFile(troublesomePullIdsLoc,
        JSON.stringify(troublesomePullIds)).done();

    for (var i = 0; i < this.children.length; ++i) {
      this.children[i].send(message);
    }
    break;
  case 'ids':
    child.send({code: 'ids', ids: this.getPullIds(message.count)});
    if (this.lastPullId > maxPullId) {
      child.stop();
    }
    break;
  case 'exit':
    this.completions[child.index] = null;
    var past = this.pastCompletions[child.index];
    this.pastCompletions[child.index] = (past ? past : 0) + message.completions;
    this.completeUpTos[child.index] = message.completeUpTo;
    break;
  case 'status':
    this.completions[child.index] = message.completions;
    if (message.completeUpTo) {
      this.completeUpTos[child.index] = message.completeUpTo;
    }
    break;
  default:
    Parent.prototype.onChildMessage.apply(this, arguments);
  }
};

Main.prototype.onChildStart = function onChildStart(child) {
  child.send({
    code: 'init',
    troublesomePullIds: troublesomePullIds,
    callbackPath: this.callbackPath,
    id: child.index,
  });
};

Main.prototype.updateDisplay = function updateDisplay(sync) {
  var disp = this.calculateAvgRate().toFixed(0)
      // http://stackoverflow.com/a/2901298/1433127
      .replace(/\B(?=(\d{3})+(?!\d))/g, ",");
  var out = disp + new Array(15 - disp.length + 1).join(' ') + 'pulls/min';

  var disp = (this.calculateTotalCompletions() + '')
      .replace(/\B(?=(\d{3})+(?!\d))/g, ",");
  out += '\n' + disp + new Array(15 - disp.length + 1).join(' ') +
      'pulls';

  var disp = (this.calculateComleteUpto() + '');
  out += '\n' + disp + new Array(15 - disp.length + 1).join(' ') +
      'completeUpTo';

  if (sync) {
    fs.writeFileSync('status.txt', out, {flag: 'w'});
  } else {
    promiseWriteFile('status.txt', out, {flag: 'w'}).done();
  }
};

Main.prototype.getPullIds = function getPullIds(count) {
  var ids = [];
  while (ids.length < count && this.lastPullId <= maxPullId) {
    ids.push(this.lastPullId++);
  }
  return ids;
};
