/**
 * var iterate = require('lib/IterateOverThings/Parent');
 * iterate({
 *   pGetMoreThings(count): a function returning a promise to get more things.
 *       resolve to null when all things have been done
 *       once it resolves to null, it should always resolve to null
 *   pWorkerInit(id): (optional) a function returning a promise which
 *       accepts the child id and resolves to an object (msg) passed to the
 *       worker on init.
 *   pStop: (optional) any cleanup for the parent process
 *   workerCount: how many worker processes? default: 4
 *   maxActiveThingsPerWorker: how many things should a worker process be
 *       allowed to do at once? default: 200
 *   maxCompleteThingsPerWorker: how many things should a worker process be
 *       allowed to do before being replaced with a new one? default: 1000
 *   workerPath: the absolute file path to worker module for processing things
 * });
 *
 * the module exports: {
 *   pInit(id, msg): (optional) accepts the init object passed from the parent
 *       (see pWorkerStart) resolves when done initializing the worker and ready
 *       to process things.
 *   pProcess(thing): accept a thing and do work with it. Resolve when done.
 *   pStop(): (optional) cleanup, this working process is restarting/stopping.
 * }
 *
 * From wherever you run the file, you can `watch -n 1 cat status.txt`
 * to see the status of the iteration.
 */

module.exports = function (options) {
  var main = new Main(options);
  main.run();
};

var Parent = require('../Parent'),
    Promise = require('promise'),
    _ = require('lodash'),
    fs = require('fs'),
    workerPath = __dirname + '/Worker.js',
    debug = require('debug')('IterateOverThings:Parent'),
    promiseWriteFile = Promise.denodeify(fs.writeFile.bind(fs));

function Main(options) {
  Parent.call(this, options.workerCount || 4, workerPath);
  this.options = options;
  this.pastCompletions = [];
  this.completions = [];
  this.startTime = Date.now();
  this.intervals = [];
  this.intervals.push(setInterval(this.updateDisplay.bind(this), 1000));
  this.thingProcessingWorkerPath = options.workerPath;
  this.pGetMoreThings = options.pGetMoreThings;

  process.env.maxActiveThingsPerWorker = options.maxActiveThingsPerWorker;
  process.env.maxCompleteThingsPerWorker = options.maxCompleteThingsPerWorker;
}

Main.prototype = Object.create(Parent.prototype);

Main.prototype.onStop = function onStop() {
  _.each(this.intervals, clearInterval);
  this.updateDisplay(true);
  Parent.prototype.onStop.apply(this, arguments);
  if (this.options.pStop) {
    Promise.resolve()
    .then(this.options.pStop)
    .done();
  }
};

Main.prototype.calculateTotalCompletions =
    function calculateTotalCompletions() {
  var completions = _.reduce(this.pastCompletions, sum, 0);
  completions += _.reduce(this.completions, sum, 0);

  return completions;

  function sum(total, num) {
    return total + (num % 1 !== 0 ? 0 : num);
  };
}

Main.prototype.calculateAvgRate = function calculateAverageRate() {
  return this.calculateTotalCompletions() /
      (Date.now() - this.startTime) * 1000 * 60;
};

Main.prototype.onChildMessage = function onChildMessage(child, message) {
  switch (message.code) {
  case 'things':	  
    this.pGetMoreThings(message.count)
    .done(function(things) {
      if (things === null) {
        child.stop();
      } else {
        child.send({code: 'things', things: things});
      }
    });
    break;
  case 'exit':
    this.completions[child.index] = null;
    var past = this.pastCompletions[child.index];
    this.pastCompletions[child.index] = (past ? past : 0) + message.completions;
    break;
  case 'status':
    this.completions[child.index] = message.completions;
    break;
  default:
    Parent.prototype.onChildMessage.apply(this, arguments);
  }
};

Main.prototype.onChildStart = function onChildStart(child) {	
  child.send({
    code: 'init',
    workerPath: this.thingProcessingWorkerPath,
    id: child.index,
    msg: "Success!!!" //TODO: Make msg something actually useful.
  });
};

Main.prototype.updateDisplay = function updateDisplay(sync) {
  var disp = this.calculateAvgRate().toFixed(0)
      // http://stackoverflow.com/a/2901298/1433127
      .replace(/\B(?=(\d{3})+(?!\d))/g, ",");
  var out = disp + new Array(15 - disp.length + 1).join(' ') + 'pulls/min';

  var disp = (this.calculateTotalCompletions() + '')
      .replace(/\B(?=(\d{3})+(?!\d))/g, ",");
  out += '\n' + disp + new Array(15 - disp.length + 1).join(' ') +
      'pulls';

  if (sync) {
    fs.writeFileSync('status.txt', out, {flag: 'w'});
  } else {
    promiseWriteFile('status.txt', out, {flag: 'w'}).done();
  }
};
