var Promise = require('promise'),
    _ = require('lodash'),
    child_process = require('child_process');

var debug = {
  parent: require('debug')('MultiProcess:Parent'),
  child: {
    restarted: require('debug')('MultiProcess:Parent:Child:Restart'),
    stopped: require('debug')('MultiProcess:Parent:Child:Stop'),
    error: require('debug')('MultiProcess:Parent:Child:Error'),
    create: require('debug')('MultiProcess:Parent:Child:Create'),
  }
};

module.exports = Parent;

/**
 * Perform long-duration parallelizable tasks in worker processes.
 *
 * to child:
 *  'stop' finish work then call 'stop'
 *
 * from child:
 *  'stop' kill this child
 *  'restart' kill & replace this child
 *
 * SIGINT (crtl+c) once to stop
 * SIGINT twice to kill
 */
function Parent(processes, workerPath) {
  this.processes = processes;
  this.workerPath = workerPath;
  this.children = [];

  var self = this;
  var sigint = false;
  process.on('SIGINT', function() {
    if (!sigint) {
      sigint = true;
      self.stop();
      console.log('SIGINT again to kill');
    } else {
      process.exit();
    }
  });
  process.on('exit', function() {
    self.kill();
  });
  this.running = false;
}

Parent.prototype.run = function run() {
  if (this.running) {
    throw new Error('already running!');
  }

  this.running = true;
  for (var i = 0; i < this.processes; ++i) {
    this.newChild();
  }
};

Parent.prototype.newChild = function newChild(index) {
  var index = index === undefined ? this.children.length : index;

  var nodeChild = child_process.fork(this.workerPath, {env: process.env});
  var child = new Child(nodeChild, this, index);
  this.children[index] = child;
  debug.child.create('%s', index);

  var parent = this;
  nodeChild.on('message', function(message) {
    parent.onChildMessage(child, message);
  });
  nodeChild.on('disconnect', function() {
    parent.onChildDisconnect(child);
  });
  nodeChild.on('error', function(e) {
    parent.onChildError(child, e);
  });

  this.onChildStart(child);
};

Parent.prototype.onChildStart = function onChildStart(child) {
};

Parent.prototype.onChildMessage = function onChildStop(child, message) {
  switch (message.code) {
  case 'error':
	for (var p in message.error) 
		console.log(p + ": " + message.error[p])
	for (var p in message.error.body) 
		console.log(p + ": " + message.error.body[p])
    var err = new Error(message.error._message);
    err.stack = message.error._stack;
    
	if (message.error.code == "ETIMEDOUT") {
		this.onChildMessage(child, {code: 'restart'});
	} else {
		this.onChildError(child, err);
	}
    break;
  case 'stop':
    debug.child.stopped(child.index);
    child.kill();
    if (!_.findWhere(this.children, {running: true})) {
      this.running = false;
      this.onStop();
    }
    break;
  case 'restart':
    debug.child.restarted(child.index);
    child.kill();
    this.newChild(child.index);
    break;
  default:
    throw new Error('(' + message.code + ') unrecognized code from child #' +
        child.index);
  }
};

Parent.prototype.onChildError = function onChildError(child, error) {
  debug.child.error(child.index);
  throw error;
};

Parent.prototype.onChildDisconnect = function onChildDisconnect(child) {
  // this was expected
  if (!child.running || child.stopped) {
    return;
  }

  throw new Error('child #' + child.index + ' unexpectedly disconnected');
};

Parent.prototype.onStop = function onStop() {
};

Parent.prototype.kill = function kill() {
  if (this.running && !this.killHappend) {
    this.killHappened = true;
    debug.parent('killing all children!');
    for (var i = 0; i < this.children.length; ++i) {
      this.children[i].kill();
    }
    this.running = false;
    this.onStop();
  }
};

Parent.prototype.stop = function stop() {
  if (this.running && !this.stopHappened) {
    this.stopHappened = true;
    debug.parent('stopping all children!');
    for (var i = 0; i < this.children.length; ++i) {
      if (!this.children[i].stopped) {
        this.children[i].stop();
      }
    }
    this.running = false;
    this.onStop();
  }
};

/******************************************************************************/
// Parent's view of a child. Wrap's node's child
function Child(nodeChild, parent, index) {
  this.nodeChild = nodeChild;
  this.parent = parent;
  this.index = index;
  this.running = true;
  this.stopped = false;
}

Child.prototype.send = function send(obj) {
  this.nodeChild.send(obj);
};

Child.prototype.kill = function kill() {
  if (this.running) {
    this.nodeChild.kill();
    this.running = false;
  }
};

Child.prototype.stop = function stop() {
  if (this.stopped) {
    throw new Error('child #' + this.index + ' already stopped!');
  }

  this.stopped = true;
  this.nodeChild.send({code: 'stop'});
};
