require("MatchIt")
require("binom")
require("ggplot2")
require("scales")
require("gridExtra")
require("xtable")
require("plyr")
require("RMySQL")

source('psm_125/dataset.R')
source('Publication - Conference - Gender Bias in Pull Requests/r/common.R')

compareMerges <- function(name, data, condition){
	
	ar <- function(data){
		weighted.mean(ifelse(data$merge_state=='Merged',1,0),data$weights)
	}
	
	ct <- function(data){
		nrow(data[which(data$weights>0),])
	}
	
	summary <- function(data){
		merge_rate <- ar(data)
		pulls <- ct(data)
		int <- binom.confint(merge_rate*pulls, pulls, 0.95, 'exact');
		return(data.frame(
				   name = name,
				   gender = ifelse(data$gender[[1]]==0,"Men","Women"),
				   merge_rate = merge_rate,
				   pulls = pulls,
				   lower = int$lower,
				   upper = int$upper
				))
	}
	
	men <- summary(data[which(data$gender==0 & condition),])
	women <- summary(data[which(data$gender==1 & condition),])
	
	women$p <- men$p <- chisq(men,women)$p.value
	
	return(rbind(women,men))
}

chisq <- function(dfA, dfB){
	
	a <- c(dfA$pulls*dfA$merge_rate,dfA$pulls*(1-dfA$merge_rate))
	b <- c(dfB$pulls*dfB$merge_rate,dfB$pulls*(1-dfB$merge_rate))
	
	return(chisq.test(rbind(a,b)))
}

returnOrCreate <- function(d,match_formula,con, pow){

	if (missing(d)){
		raw_d <- read_raw_dataset(con)
		d = load_dataset(raw_d, pow)
		m.out <- matchit(match_formula, data=d, method='exact')
		d$weights = m.out$weights
	} 
	return(d)
}

#d has to be defined as an exact matched dataframe with $weights
psm.overall <- function(d,con){
	
	d <- returnOrCreate(d,gender ~ 
						lines_added + lines_deleted +
						commit_count + files_changed +
						creator_type + references_from + 
						license + creator_pull_index + 
						extension + identicon, con, pow = 2)
	
	result <- 				compareMerges("all",d,TRUE)

	result <- rbind(result,	compareMerges("Outsiders",d,d$creator_type=='Outsider'))
	
	result <- rbind(result,	compareMerges("OSS",d,d$license=='Open Source'))
		
	result$name <- paste(result$name,ifelse(result$p < 0.05,'*',''))	
	
	po <- plot(result,"matchOverallTable")
	pdfOn("overall-match.pdf")
	print(x=po)
	dev.off()
	
	tc <- data.frame()	
	for(ext in c('.js', '.rb', '.py', '.java', '.php', '.cpp', '.cs', '.c', '.go', '.m')){
		tc <- rbind(tc, compareMerges(ext,d,d$extension_raw==ext))
	}	
	
	ntc <- data.frame()	
	for(ext in c('.md', '.html', '.xml', '.json', '.txt', '.yml', '.rst', '.markdown', '.podspec')){
		ntc <- rbind(ntc,	compareMerges(ext,d,d$extension_raw==ext))
	}
	
	#### Add and adjust p values
	
	ps <- rbind(unique(tc[,c("name","p")]),unique(ntc[,c("name","p")]))
	ps$p.adj <- p.adjust(ps$p,"BH")
	
	tc <- join(subset(tc,select=-p),  ps,by="name")
	ntc <- join(subset(ntc,select=-p),ps,by="name")
	
	tc$name <- 	paste(tc$name,ifelse(tc$p.adj < 0.05,'*',''))
	ntc$name <- paste(ntc$name,ifelse(ntc$p.adj < 0.05,'*',''))	
	
	####
	
	ptc <- plot(tc,"matchExtTCTable")
	pntc <- plot(ntc,"matchExtNTCTable")
	pdfOn("accepts_by_language-match.pdf")
	grid.arrange(ptc, pntc, nrow=2)
	dev.off()
	
	result <- data.frame()
	for(n in c(0:6)){#limit 128, where we stopped preveviously due to few women at that level
		result <- rbind(result,compareMerges(paste0(2^n,"-",2^(n+1)-1," PRs"),d,d$creator_pull_index==n))
	}
	
	adj <- p.adjust(result[c(TRUE,FALSE),]$p,"BH")
	result$p.adj <- c(rbind(adj,adj))

	result$name <- paste(result$name,ifelse(result$p.adj < 0.05,'*',''))	
	
	pt <- plot(result,"matchTimeTable")
	pdfOn("accepts_over_time-match.pdf")
	print(x=pt)
	dev.off()
	
	#follow up analysis
	first.pulls(d)
	
	return(result)
}

pdfOn <- function(file){
	pdf(paste0("Publication - Conference - Gender Bias in Pull Requests/figs/",file))
}

plot <- function(result,tableName){
	
	resultTable <- xtable(result,auto=TRUE)	
	newcommand(tableName,
			print(resultTable,
					hline.after=c(0),		
					include.rownames=FALSE,
					print.results=FALSE,
					comment=FALSE,
					floating=FALSE))
	
	#necessary for mainintain order of names in plot
	result$name <- factor(result$name, levels=unique(result$name))
	ptc <- ggplot(data=result, aes(x=name, ymin=lower, ymax=upper, fill=gender,
							lower=merge_rate, upper=merge_rate, middle=merge_rate, color=gender)) +
			geom_boxplot(stat='identity', position=position_dodge(0.05)) +
			scale_fill_manual(labels = c('Women','Men'), values=c('#ffbf80','#660066'), name='') +  
			scale_colour_manual(labels = c('Women','Men'), values=c('#ffbf80','#660066'), name='') +
			xlab("") +
			ylab("Acceptance Rate") +
			y_ar_scale() +
			theme_bw() +
			theme(legend.text = element_text(size = 12), legend.position = c(0.8,0.95), legend.direction = "horizontal", legend.key = element_rect(color = NA)) +
			theme(axis.text.x = element_text(angle = -38, hjust = 0)) +
			theme(axis.line = element_line(size = 0.4, colour = "black", linetype = "solid")) +
			theme(panel.border = element_rect(linetype = "blank")) +
			theme(panel.grid = element_line(linetype = "blank"),
					panel.grid.major.y = element_line(linetype = 'dotted', color="#cccccc", size=0.3),
					panel.grid.minor.y = element_line(linetype = 'dotted', color="#cccccc", size=0.2),
					panel.grid.major.x = element_line(linetype = 'solid', color="#eeeeee", size=15))
	
	return(ptc)
}

psm.identifiable <- function(d,con,pow){
	
	d <- returnOrCreate(d,gender ~ 
					lines_added + lines_deleted +
					commit_count + files_changed +
					creator_type + references_from + 
					license + creator_pull_index + 
					extension + identifiable,con,pow)

	di <- compareMerges("Gender-Neutral",d,d$creator_type!='Outsider' & d$identifiable=='neutral')
	di <- rbind(di,	compareMerges("Gendered",d,d$creator_type!='Outsider' & d$identifiable=='gendered'))
	
	di$name <- paste(di$name,ifelse(di$p < 0.05,'*',''))	
	names(di)[names(di)=="name"] <- "identifiable"
	
	do <- compareMerges("Gender-Neutral",d,d$creator_type=='Outsider' & d$identifiable=='neutral')
	do <- rbind(do,	compareMerges("Gendered",d,d$creator_type=='Outsider' & d$identifiable=='gendered'))

	do$name <- paste(do$name,ifelse(do$p < 0.05,'*',''))	
	names(do)[names(do)=="name"] <- "identifiable"
	
	result <- rbind(di,do)
	
	# print the underpowered table if we're still doing log2
	resultTable <- xtable(result,auto=TRUE)	
	newcommand(paste0("matchIdentifiableTable",ifelse(pow==2,'Underpowered','')),
			print(resultTable,
					hline.after=c(0),		
					include.rownames=FALSE,
					print.results=FALSE,
					comment=FALSE,
					floating=FALSE))
	
	if(pow!=2){
		
		source("perceived_gender_36/second.R")
		plot.identifiable(di,do,"Publication - Conference - Gender Bias in Pull Requests/figs/gendered-match.pdf")
		
		who <- function(gender,name){
			return(result[which(result$gender==gender & result$identifiable==name),])
		}
	
		wog <- who('Women','Gendered *')
		mog <- who('Men','Gendered *')
		
		newcommandPercent("matchIdChange",abs(mog$merge_rate-wog$merge_rate))
		newcommand("matchIdChangeSig",chiSqFormat(chisq(mog,wog),.01))
	}
}

first.pulls <- function(d){
	m<-dbDriver("MySQL");
	con<-dbConnect(m,user=Sys.getenv('MYSQL_USER'),
			password=Sys.getenv('MYSQL_PASSWORD'),
			host=Sys.getenv('MYSQL_SERVER'),
			port=Sys.getenv('MYSQL_SERVER_PORT'),
			dbname='ghtorrent', client.flag = CLIENT_MULTI_STATEMENTS);

	pull.counts <- dbGetQuery(con, 
		"select creator_id as user,min(pull_request_id) as pull_id,
		max(creator_pull_index) as pulls from g_pull_closes2 group by creator_id")
	
	dbDisconnect(con)
	
	firstTimerPulls <- d$pull_request_id %in% pull.counts[pull.counts$pulls==1,]$pull_id
	
	result <- compareMerges("One-Timers",d,firstTimerPulls)
	result <- rbind(result,compareMerges("Regulars' First",d,!firstTimerPulls & d$creator_pull_index==1))
	result <- rbind(result,compareMerges("Regulars' Rest",d, !firstTimerPulls & d$creator_pull_index!=1))
	
	result$name <- paste(result$name,ifelse(result$p < 0.05,'*',''))	
	
	po <- plot(result,"matchTimeBreakdownTable")
	pdfOn("accepts_over_time_breakdown-match.pdf")
	print(x=po)
	dev.off()
	
	return(result)
}	